#include <iostream>
#include <memory>
#include <vector>

struct tree {
	int value;
	std::unique_ptr<tree> left, right;
};

class tree_visitor_base {
	virtual void init() {}
	virtual void visit_node(const tree&) {}

	void rec_traverse(const tree* t, int order) {
		if (t == nullptr)
			return;

		if (order == 0)
			visit_node(*t);

		rec_traverse(t->left.get(), order);

		if (order == 1)
			visit_node(*t);

		rec_traverse(t->right.get(), order);

		if (order == 2)
			visit_node(*t);
	}

	void traverse(const tree& t, int order) {
		init();
		rec_traverse(&t, order);
	}

public:
	void preorder(const tree& t) { traverse(t, 0); }
	void inorder(const tree& t) { traverse(t, 1); }
	void postorder(const tree& t) { traverse(t, 2); }
};

class tree_collector : public tree_visitor_base {
	std::vector<int> values;

	void init() override { values.clear(); }

	void visit_node(const tree& t) override {
		values.push_back(t.value);
	}

public:
	// supposed to be called once after running (pre/post/in)order
	std::vector<int> get() { return std::move(values); }
};

void print_vec(const auto& vec) {
	for (auto& elem : vec) {
		std::cout << ' ' << elem;
	}
	std::cout << '\n';
}

int main() {
	tree t = { 7,
		std::make_unique<tree>(5,
			std::make_unique<tree>(1),
			std::make_unique<tree>(3)
		),
		std::make_unique<tree>(2,
			std::make_unique<tree>(4),
			nullptr
		)
	};

	tree_collector tc;
	tc.preorder(t);
	print_vec(tc.get());

	tc.inorder(t);
	print_vec(tc.get());

	tc.postorder(t);
	print_vec(tc.get());
}
